// Proof-of-concept bitslice implementation of the SHA256 compression function.
//
// Written by Alain Espinosa <alainesp at gmail.com> in 2015, and placed in the
// public domain. There's absolutely no warranty. Although this is not required
// formally, due credit will be appreciated if you re-use this code or concepts.
            

//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
// Bitslice SHA256 format
//;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
#define state r0
#define W     r1
#define bs_A  r2
#define REG_BYTE_SIZE 16


#define step_H     r3
#define step_E     r4
#define step_G     r5
#define step_F     r6
#define step_const r7
#define W_ptr      r8
#define step_D     r9
#define step_A     r10
#define step_B     r11
#define step_C     r12

// Sums carries
#define carries_bs    q4
#define carries_RE    q5
#define carries_const q6
#define carries_W     q7

#define carries_D     q8

#define carries_maj   q9
#define carries_RA    q10
#define ones          q11
#define t0			  q12
#define t1			  q13
//#define t2			  q14
//#define t3			  q15

#define sum0   q0
#define sum1   q1
#define reg_H  q2
#define reg_H1 q3

#define BS_SHA256_STEP_AVX2_UNROLL_0 \
		/* Calculate BITSELECT*/\
		vld1.u32 {t0}, [step_G:128]!;\
		vld1.u32 {t1}, [step_F:128]!;\
		vld1.u32 {sum1}, [step_E:128];\
		vbsl sum1, t1, t0;\
		/* Sum BITSELECT*/\
		vld1.u32 {sum0}, [step_H:128];\
		veor reg_H     , sum1, sum0;\
		vand carries_bs, sum1, sum0;\
		/* Sum RE*/\
		add r14, step_E, #(((32 - 26 + 0) & 31)*REG_BYTE_SIZE);\
		vld1.u32 {sum0}, [r14:128];\
		add r14, step_E, #(((32 - 21 + 0) & 31)*REG_BYTE_SIZE);\
		vld1.u32 {t0}, [r14:128];\
		add r14, step_E, #(((32 -  7 + 0) & 31)*REG_BYTE_SIZE);\
		vld1.u32 {t1}, [r14:128];\
		veor sum0, sum0, t0;\
		veor sum0, sum0, t1;\
		veor   reg_H1  , sum0, reg_H;\
		vand carries_RE, sum0, reg_H;\
		/* Sum CONST*/\
		tst step_const, #1;\
		beq zero_const0;\
			vmov  carries_const, reg_H1;\
			veor reg_H, reg_H1, ones;\
			b const_end0;\
zero_const0:\
			vmov reg_H, reg_H1;\
			vmov.u32 carries_const, #0;\
const_end0:\
		/* Sum W*/\
		vld1.u32 {sum0}, [W_ptr:128]!;\
		veor reg_H1   , sum0 , reg_H;\
		vand carries_W, sum0 , reg_H;\
\
		/* Sum D*/\
		vld1.u32 {sum0}, [step_D:128];\
		veor   t0     , reg_H1, sum0;\
		vst1.u32  {t0}, [step_D:128]!;\
		vand carries_D, reg_H1, sum0;\
\
		/* Calculate MAJ*/\
		vld1.u32 { t0 }, [step_A:128]!;\
		vld1.u32 { t1 }, [step_B:128]!;\
		vld1.u32 {sum0}, [step_C:128]!;\
		veor sum0, t1, sum0;\
		vbsl sum0, t0, t1;\
		/* Sum MAJ*/\
		veor   reg_H    ,  sum0 , reg_H1;\
		vand carries_maj,  sum0 , reg_H1;\
		/* Sum RA*/\
		vld1.u32 {sum0}, [r0:128]!;\
		vld1.u32 { t0 }, [r1:128]!;\
		vld1.u32 { t1 }, [r2:128]!;\
		veor sum0, sum0, t0;\
		veor sum0, sum0, t1;\
		veor       t0  , reg_H, sum0;\
		vst1.u32  {t0} , [step_H:128]!;\
		vand carries_RA, reg_H, sum0;

#define BS_SHA256_STEP_AVX2_UNROLL(i) \
		/* Calculate BITSELECT*/\
		vld1.u32 {t0}, [step_G:128]!;\
		vld1.u32 {t1}, [step_F:128]!;\
		add r14, step_E, #(i*REG_BYTE_SIZE);\
		vld1.u32 {sum1}, [r14:128];\
		vbsl sum1, t1, t0;\
		/* Sum BITSELECT*/\
		vld1.u32 {sum0}, [step_H:128];\
		veor carries_bs, sum0, carries_bs;\
		veor reg_H     , sum1, carries_bs;\
		vbsl carries_bs, sum1, sum0;\
		/* Sum RE*/\
		add r14, step_E, #(((32 - 26 + i) & 31)*REG_BYTE_SIZE);\
		vld1.u32 {sum0}, [r14:128];\
		add r14, step_E, #(((32 - 21 + i) & 31)*REG_BYTE_SIZE);\
		vld1.u32 {t0}, [r14:128];\
		add r14, step_E, #(((32 -  7 + i) & 31)*REG_BYTE_SIZE);\
		vld1.u32 {t1}, [r14:128];\
		veor sum0, sum0, t0;\
		veor sum0, sum0, t1;\
		veor carries_RE, carries_RE, reg_H;\
		veor reg_H1    , carries_RE, sum0 ;\
		vbsl carries_RE,    sum0   , reg_H;\
		/* Sum CONST*/\
		veor reg_H, reg_H1, carries_const;\
		tst step_const, #(1<<i);\
		beq zero_const ## i ;\
			vorr  carries_const, reg_H, reg_H1;\
			veor reg_H, reg_H, ones;\
			b const_end ## i ;\
zero_const ## i :\
			vand carries_const, reg_H1, carries_const;\
const_end ## i :\
		/* Sum W*/\
		vld1.u32 {sum0}, [W_ptr:128]!;\
		veor carries_W, reg_H, carries_W;\
		veor reg_H1   , sum0 , carries_W;\
		vbsl carries_W, sum0 , reg_H;\
\
		/* Sum D*/\
		vld1.u32 {sum0}, [step_D:128];\
		veor carries_D , sum0, carries_D;\
		veor t0, reg_H1, carries_D;\
		vst1.u32  {t0} , [step_D:128]!;\
		vbsl carries_D , reg_H1, sum0;\
\
		/* Calculate MAJ*/\
		vld1.u32 { t0 }, [step_A:128]!;\
		vld1.u32 { t1 }, [step_B:128]!;\
		vld1.u32 {sum0}, [step_C:128]!;\
		veor sum0, t1, sum0;\
		vbsl sum0, t0, t1;\
		/* Sum MAJ*/\
		veor carries_maj, reg_H1, carries_maj;\
		veor   reg_H    ,  sum0 , carries_maj;\
		vbsl carries_maj,  sum0 , reg_H1;\
		/* Sum RA*/\
		vld1.u32 {sum0}, [r0:128]!;\
		vld1.u32 { t0 }, [r1:128]!;\
		vld1.u32 { t1 }, [r2:128]!;\
		veor sum0, sum0, t0;\
		veor sum0, sum0, t1;\
		veor carries_RA, sum0, carries_RA;\
		veor t0       , reg_H, carries_RA;\
		vst1.u32 {t0}, [step_H:128]!;\
		vbsl carries_RA, reg_H, sum0;
		
#define BS_SHA256_STEP_AVX2_UNROLL_LAST \
		/* Calculate BITSELECT*/\
		vld1.u32 {t0}, [step_G:128]!;\
		vld1.u32 {t1}, [step_F:128]!;\
		add r14, step_E, #(31*REG_BYTE_SIZE);\
		vld1.u32 {sum1}, [r14:128];\
		vbsl sum1, t1, t0;\
		/* Sum BITSELECT*/\
		vld1.u32 {sum0}, [step_H:128];\
		veor carries_bs, sum0, carries_bs;\
		veor reg_H     , sum1, carries_bs;\
		/* Sum RE*/\
		add r14, step_E, #(((32 - 26 + 31) & 31)*REG_BYTE_SIZE);\
		vld1.u32 {sum0}, [r14:128];\
		add r14, step_E, #(((32 - 21 + 31) & 31)*REG_BYTE_SIZE);\
		vld1.u32 {t0}, [r14:128];\
		add r14, step_E, #(((32 -  7 + 31) & 31)*REG_BYTE_SIZE);\
		vld1.u32 {t1}, [r14:128];\
		veor sum0, sum0, t0;\
		veor sum0, sum0, t1;\
		veor carries_RE, carries_RE, reg_H;\
		veor reg_H1    , carries_RE, sum0 ;\
		/* Sum CONST*/\
		veor reg_H, reg_H1, carries_const;\
		tst step_const, #(1<<31);\
		beq zero_const31;\
			veor reg_H, reg_H, ones;\		
zero_const31:\
		/* Sum W*/\
		vld1.u32 {sum0}, [W_ptr:128]!;\
		veor carries_W, reg_H, carries_W;\
		veor reg_H1   , sum0 , carries_W;\
\
		/* Sum D*/\
		vld1.u32 {sum0}, [step_D:128];\
		veor carries_D , sum0, carries_D;\
		veor t0, reg_H1, carries_D;\
		vst1.u32  {t0} , [step_D:128]!;\
\
		/* Calculate MAJ*/\
		vld1.u32 { t0 }, [step_A:128]!;\
		vld1.u32 { t1 }, [step_B:128]!;\
		vld1.u32 {sum0}, [step_C:128]!;\
		veor sum0, t1, sum0;\
		vbsl sum0, t0, t1;\
		/* Sum MAJ*/\
		veor carries_maj, reg_H1, carries_maj;\
		veor   reg_H    ,  sum0 , carries_maj;\
		/* Sum RA*/\
		vld1.u32 {sum0}, [r0:128]!;\
		vld1.u32 { t0 }, [r1:128]!;\
		vld1.u32 { t1 }, [r2:128]!;\
		veor sum0, sum0, t0;\
		veor sum0, sum0, t1;\
		veor carries_RA, sum0, carries_RA;\
		veor t0       , reg_H, carries_RA;\
		vst1.u32 {t0}, [step_H:128]!;

	.text
	.align	2
	.global	bs_sha256_step_neon
	.type	bs_sha256_step_neon, %function
bs_sha256_step_neon:
	vmov.u32 d30[0], r14
	vmov.u32 d30[1], r0
	vmov.u32 d31[0], r1
	vmov.u32 d31[1], r2
	
	add r0, step_A, #( 2*REG_BYTE_SIZE)
	add r1, step_A, #(13*REG_BYTE_SIZE)
	add r2, step_A, #(22*REG_BYTE_SIZE)
	BS_SHA256_STEP_AVX2_UNROLL_0
	BS_SHA256_STEP_AVX2_UNROLL( 1)
	BS_SHA256_STEP_AVX2_UNROLL( 2)
	BS_SHA256_STEP_AVX2_UNROLL( 3)
	BS_SHA256_STEP_AVX2_UNROLL( 4)
	BS_SHA256_STEP_AVX2_UNROLL( 5)
	BS_SHA256_STEP_AVX2_UNROLL( 6)
	BS_SHA256_STEP_AVX2_UNROLL( 7)
	BS_SHA256_STEP_AVX2_UNROLL( 8)
	BS_SHA256_STEP_AVX2_UNROLL( 9)
	sub r2, r2, #(32*REG_BYTE_SIZE)
	BS_SHA256_STEP_AVX2_UNROLL(10)
	BS_SHA256_STEP_AVX2_UNROLL(11)
	BS_SHA256_STEP_AVX2_UNROLL(12)
	BS_SHA256_STEP_AVX2_UNROLL(13)
	BS_SHA256_STEP_AVX2_UNROLL(14)
	BS_SHA256_STEP_AVX2_UNROLL(15)
	BS_SHA256_STEP_AVX2_UNROLL(16)
	BS_SHA256_STEP_AVX2_UNROLL(17)
	BS_SHA256_STEP_AVX2_UNROLL(18)
	sub r1, r1, #(32*REG_BYTE_SIZE)
	BS_SHA256_STEP_AVX2_UNROLL(19)
	BS_SHA256_STEP_AVX2_UNROLL(20)
	BS_SHA256_STEP_AVX2_UNROLL(21)
	BS_SHA256_STEP_AVX2_UNROLL(22)
	BS_SHA256_STEP_AVX2_UNROLL(23)
	BS_SHA256_STEP_AVX2_UNROLL(24)
	BS_SHA256_STEP_AVX2_UNROLL(25)
	BS_SHA256_STEP_AVX2_UNROLL(26)
	BS_SHA256_STEP_AVX2_UNROLL(27)
	BS_SHA256_STEP_AVX2_UNROLL(28)
	BS_SHA256_STEP_AVX2_UNROLL(29)
	sub r0, r0, #(32*REG_BYTE_SIZE)
	BS_SHA256_STEP_AVX2_UNROLL(30)
	BS_SHA256_STEP_AVX2_UNROLL_LAST
	
	vmov.u32 r14, d30[0]
	vmov.u32 r0 , d30[1] 
	vmov.u32 r1 , d31[0] 
	vmov.u32 r2 , d31[1] 
	
	bx lr


// Sums carries
#define carries_R1  q4
#define carries_sum q5
#define carries_R0  q6

#define sum0   q0
#define sum1   q1
#define reg_W  q2
#define reg_W1 q3

#define r1_index_0 r3
#define r1_index_1 r4
#define r1_index_2 r5
#define r0_index_0 r6
#define r0_index_1 r7
#define r0_index_2 r8

#define r_index    r9
#define r1_index   r10
#define rsum_index r11
#define r0_index   r12

#define BS_SHA256_RW_R1_INDEX_2_0	vld1.u32 {t1}, [r1_index_2:128]!;veor sum1, sum1, t1;
#define BS_SHA256_RW_R1_INDEX_2_1	vld1.u32 {t1}, [r1_index_2:128]!;veor sum1, sum1, t1;
#define BS_SHA256_RW_R1_INDEX_2_2	vld1.u32 {t1}, [r1_index_2:128]!;veor sum1, sum1, t1;
#define BS_SHA256_RW_R1_INDEX_2_3	vld1.u32 {t1}, [r1_index_2:128]!;veor sum1, sum1, t1;
#define BS_SHA256_RW_R1_INDEX_2_4	vld1.u32 {t1}, [r1_index_2:128]!;veor sum1, sum1, t1;
#define BS_SHA256_RW_R1_INDEX_2_5	vld1.u32 {t1}, [r1_index_2:128]!;veor sum1, sum1, t1;
#define BS_SHA256_RW_R1_INDEX_2_6	vld1.u32 {t1}, [r1_index_2:128]!;veor sum1, sum1, t1;
#define BS_SHA256_RW_R1_INDEX_2_7	vld1.u32 {t1}, [r1_index_2:128]!;veor sum1, sum1, t1;
#define BS_SHA256_RW_R1_INDEX_2_8	vld1.u32 {t1}, [r1_index_2:128]!;veor sum1, sum1, t1;
#define BS_SHA256_RW_R1_INDEX_2_9	vld1.u32 {t1}, [r1_index_2:128]!;veor sum1, sum1, t1;
#define BS_SHA256_RW_R1_INDEX_2_10	vld1.u32 {t1}, [r1_index_2:128]!;veor sum1, sum1, t1;
#define BS_SHA256_RW_R1_INDEX_2_11	vld1.u32 {t1}, [r1_index_2:128]!;veor sum1, sum1, t1;
#define BS_SHA256_RW_R1_INDEX_2_12	vld1.u32 {t1}, [r1_index_2:128]!;veor sum1, sum1, t1;
#define BS_SHA256_RW_R1_INDEX_2_13	vld1.u32 {t1}, [r1_index_2:128]!;veor sum1, sum1, t1;
#define BS_SHA256_RW_R1_INDEX_2_14	vld1.u32 {t1}, [r1_index_2:128]!;veor sum1, sum1, t1;
#define BS_SHA256_RW_R1_INDEX_2_15	vld1.u32 {t1}, [r1_index_2:128]!;veor sum1, sum1, t1;
#define BS_SHA256_RW_R1_INDEX_2_16	vld1.u32 {t1}, [r1_index_2:128]!;veor sum1, sum1, t1;
#define BS_SHA256_RW_R1_INDEX_2_17	vld1.u32 {t1}, [r1_index_2:128]!;veor sum1, sum1, t1;
#define BS_SHA256_RW_R1_INDEX_2_18	vld1.u32 {t1}, [r1_index_2:128]!;veor sum1, sum1, t1;
#define BS_SHA256_RW_R1_INDEX_2_19	vld1.u32 {t1}, [r1_index_2:128]!;veor sum1, sum1, t1;
#define BS_SHA256_RW_R1_INDEX_2_20	vld1.u32 {t1}, [r1_index_2:128]!;veor sum1, sum1, t1;
#define BS_SHA256_RW_R1_INDEX_2_21	vld1.u32 {t1}, [r1_index_2:128]!;veor sum1, sum1, t1;
#define BS_SHA256_RW_R1_INDEX_2_22
#define BS_SHA256_RW_R1_INDEX_2_23
#define BS_SHA256_RW_R1_INDEX_2_24
#define BS_SHA256_RW_R1_INDEX_2_25
#define BS_SHA256_RW_R1_INDEX_2_26
#define BS_SHA256_RW_R1_INDEX_2_27
#define BS_SHA256_RW_R1_INDEX_2_28
#define BS_SHA256_RW_R1_INDEX_2_29
#define BS_SHA256_RW_R1_INDEX_2_30

#define BS_SHA256_RW_R0_INDEX_2_0	vld1.u32 {t1}, [r0_index_2:128]!;veor sum0, sum0, t1;
#define BS_SHA256_RW_R0_INDEX_2_1	vld1.u32 {t1}, [r0_index_2:128]!;veor sum0, sum0, t1;
#define BS_SHA256_RW_R0_INDEX_2_2	vld1.u32 {t1}, [r0_index_2:128]!;veor sum0, sum0, t1;
#define BS_SHA256_RW_R0_INDEX_2_3	vld1.u32 {t1}, [r0_index_2:128]!;veor sum0, sum0, t1;
#define BS_SHA256_RW_R0_INDEX_2_4	vld1.u32 {t1}, [r0_index_2:128]!;veor sum0, sum0, t1;
#define BS_SHA256_RW_R0_INDEX_2_5	vld1.u32 {t1}, [r0_index_2:128]!;veor sum0, sum0, t1;
#define BS_SHA256_RW_R0_INDEX_2_6	vld1.u32 {t1}, [r0_index_2:128]!;veor sum0, sum0, t1;
#define BS_SHA256_RW_R0_INDEX_2_7	vld1.u32 {t1}, [r0_index_2:128]!;veor sum0, sum0, t1;
#define BS_SHA256_RW_R0_INDEX_2_8	vld1.u32 {t1}, [r0_index_2:128]!;veor sum0, sum0, t1;
#define BS_SHA256_RW_R0_INDEX_2_9	vld1.u32 {t1}, [r0_index_2:128]!;veor sum0, sum0, t1;
#define BS_SHA256_RW_R0_INDEX_2_10	vld1.u32 {t1}, [r0_index_2:128]!;veor sum0, sum0, t1;
#define BS_SHA256_RW_R0_INDEX_2_11	vld1.u32 {t1}, [r0_index_2:128]!;veor sum0, sum0, t1;
#define BS_SHA256_RW_R0_INDEX_2_12	vld1.u32 {t1}, [r0_index_2:128]!;veor sum0, sum0, t1;
#define BS_SHA256_RW_R0_INDEX_2_13	vld1.u32 {t1}, [r0_index_2:128]!;veor sum0, sum0, t1;
#define BS_SHA256_RW_R0_INDEX_2_14	vld1.u32 {t1}, [r0_index_2:128]!;veor sum0, sum0, t1;
#define BS_SHA256_RW_R0_INDEX_2_15	vld1.u32 {t1}, [r0_index_2:128]!;veor sum0, sum0, t1;
#define BS_SHA256_RW_R0_INDEX_2_16	vld1.u32 {t1}, [r0_index_2:128]!;veor sum0, sum0, t1;
#define BS_SHA256_RW_R0_INDEX_2_17	vld1.u32 {t1}, [r0_index_2:128]!;veor sum0, sum0, t1;
#define BS_SHA256_RW_R0_INDEX_2_18	vld1.u32 {t1}, [r0_index_2:128]!;veor sum0, sum0, t1;
#define BS_SHA256_RW_R0_INDEX_2_19	vld1.u32 {t1}, [r0_index_2:128]!;veor sum0, sum0, t1;
#define BS_SHA256_RW_R0_INDEX_2_20	vld1.u32 {t1}, [r0_index_2:128]!;veor sum0, sum0, t1;
#define BS_SHA256_RW_R0_INDEX_2_21	vld1.u32 {t1}, [r0_index_2:128]!;veor sum0, sum0, t1;
#define BS_SHA256_RW_R0_INDEX_2_22	vld1.u32 {t1}, [r0_index_2:128]!;veor sum0, sum0, t1;
#define BS_SHA256_RW_R0_INDEX_2_23	vld1.u32 {t1}, [r0_index_2:128]!;veor sum0, sum0, t1;
#define BS_SHA256_RW_R0_INDEX_2_24	vld1.u32 {t1}, [r0_index_2:128]!;veor sum0, sum0, t1;
#define BS_SHA256_RW_R0_INDEX_2_25	vld1.u32 {t1}, [r0_index_2:128]!;veor sum0, sum0, t1;
#define BS_SHA256_RW_R0_INDEX_2_26	vld1.u32 {t1}, [r0_index_2:128]!;veor sum0, sum0, t1;
#define BS_SHA256_RW_R0_INDEX_2_27	vld1.u32 {t1}, [r0_index_2:128]!;veor sum0, sum0, t1;
#define BS_SHA256_RW_R0_INDEX_2_28	vld1.u32 {t1}, [r0_index_2:128]!;veor sum0, sum0, t1;
#define BS_SHA256_RW_R0_INDEX_2_29
#define BS_SHA256_RW_R0_INDEX_2_30

#define BS_SHA256_RW_UNROLL_0 \
		/* Calculate R1*/\
		vld1.u32 {sum1}, [r1_index_0:128]!;\
		vld1.u32 { t0 }, [r1_index_1:128]!;\
		veor sum1, sum1, t0;\
		BS_SHA256_RW_R1_INDEX_2_0 \
		/* Sum R1*/\
		vld1.u32 {sum0}, [r_index:128];\
		veor reg_W     , sum1, sum0;\
		vand carries_R1, sum1, sum0;\
\
		/* Sum rsum_index*/\
		vld1.u32 {sum0}, [rsum_index:128]!;\
		veor     reg_W1 , sum0 , reg_W;\
		vand carries_sum, sum0 , reg_W;\
\
		/* Calculate R0*/\
		vld1.u32 {sum0}, [r0_index_0:128]!;\
		vld1.u32 { t0 }, [r0_index_1:128]!;\
		veor sum0, sum0, t0;\
		BS_SHA256_RW_R0_INDEX_2_0 \
		/* Sum R0*/\
		veor   t0     , reg_W1, sum0;\
		vst1.u32 {t0}, [r_index:128]!;\
		vand carries_R0, reg_W1, sum0;
		

#define BS_SHA256_RW_UNROLL(i) \
		/* Calculate R1*/\
		vld1.u32 {sum1}, [r1_index_0:128]!;\
		vld1.u32 { t0 }, [r1_index_1:128]!;\
		veor sum1, sum1, t0;\
		BS_SHA256_RW_R1_INDEX_2_ ## i \
		/* Sum R1*/\
		vld1.u32 {sum0}, [r_index:128];\
		veor carries_R1, sum0, carries_R1;\
		veor reg_W     , sum1, carries_R1;\
		vbsl carries_R1, sum1, sum0;\
\
		/* Sum rsum_index*/\
		vld1.u32 {sum0}, [rsum_index:128]!;\
		veor carries_sum, reg_W, carries_sum;\
		veor     reg_W1 , sum0 , carries_sum;\
		vbsl carries_sum, sum0 , reg_W;\
\
		/* Calculate R0*/\
		vld1.u32 {sum0}, [r0_index_0:128]!;\
		vld1.u32 { t0 }, [r0_index_1:128]!;\
		veor sum0, sum0, t0;\
		BS_SHA256_RW_R0_INDEX_2_ ## i \
		/* Sum R0*/\
		veor carries_R0, sum0, carries_R0;\
		veor   t0     , reg_W1, carries_R0;\
		vst1.u32 {t0}, [r_index:128]!;\
		vbsl carries_R0, reg_W1, sum0;
		
#define BS_SHA256_RW_UNROLL_LAST \
		/* Calculate R1*/\
		vld1.u32 {sum1}, [r1_index_0:128]!;\
		vld1.u32 { t0 }, [r1_index_1:128]!;\
		veor sum1, sum1, t0;\
		/* Sum R1*/\
		vld1.u32 {sum0}, [r_index:128];\
		veor carries_R1, sum0, carries_R1;\
		veor reg_W     , sum1, carries_R1;\
\
		/* Sum rsum_index*/\
		vld1.u32 {sum0}, [rsum_index:128]!;\
		veor carries_sum, reg_W, carries_sum;\
		veor     reg_W1 , sum0 , carries_sum;\
\
		/* Calculate R0*/\
		vld1.u32 {sum0}, [r0_index_0:128]!;\
		vld1.u32 { t0 }, [r0_index_1:128]!;\
		veor sum0, sum0, t0;\
		/* Sum R0*/\
		veor carries_R0, sum0, carries_R0;\
		veor   t0     , reg_W1, carries_R0;\
		vst1.u32 {t0}, [r_index:128]!;
		

	.text
	.align	2
	.global	bs_sha256_RW_neon
	.type	bs_sha256_RW_neon, %function
bs_sha256_RW_neon:

	add r1_index_0, r1_index, #(17*REG_BYTE_SIZE)
	add r1_index_1, r1_index, #(19*REG_BYTE_SIZE)	
	add r1_index_2, r1_index, #(10*REG_BYTE_SIZE)
	
	add r0_index_0, r0_index, #( 7*REG_BYTE_SIZE)
	add r0_index_1, r0_index, #(18*REG_BYTE_SIZE)
	add r0_index_2, r0_index, #( 3*REG_BYTE_SIZE)

	BS_SHA256_RW_UNROLL_0
	BS_SHA256_RW_UNROLL( 1)
	BS_SHA256_RW_UNROLL( 2)
	BS_SHA256_RW_UNROLL( 3)
	BS_SHA256_RW_UNROLL( 4)
	BS_SHA256_RW_UNROLL( 5)
	BS_SHA256_RW_UNROLL( 6)
	BS_SHA256_RW_UNROLL( 7)
	BS_SHA256_RW_UNROLL( 8)
	BS_SHA256_RW_UNROLL( 9)
	BS_SHA256_RW_UNROLL(10)
	BS_SHA256_RW_UNROLL(11)
	BS_SHA256_RW_UNROLL(12)
	mov r1_index_1, r1_index
	BS_SHA256_RW_UNROLL(13)
	mov r0_index_1, r0_index
	BS_SHA256_RW_UNROLL(14)
	mov r1_index_0, r1_index
	BS_SHA256_RW_UNROLL(15)
	BS_SHA256_RW_UNROLL(16)
	BS_SHA256_RW_UNROLL(17)
	BS_SHA256_RW_UNROLL(18)
	BS_SHA256_RW_UNROLL(19)
	BS_SHA256_RW_UNROLL(20)
	BS_SHA256_RW_UNROLL(21)
	BS_SHA256_RW_UNROLL(22)
	BS_SHA256_RW_UNROLL(23)
	BS_SHA256_RW_UNROLL(24)
	mov r0_index_0, r0_index
	BS_SHA256_RW_UNROLL(25)
	BS_SHA256_RW_UNROLL(26)
	BS_SHA256_RW_UNROLL(27)
	BS_SHA256_RW_UNROLL(28)
	BS_SHA256_RW_UNROLL(29)
	BS_SHA256_RW_UNROLL(30)
	BS_SHA256_RW_UNROLL_LAST
	
	bx lr


#define CALL_BS_SHA256_RW_AVX2(pr_index, pr1_index, prsum_index, pr0_index) \
\
	add r1_index  , W, #(pr1_index*32*REG_BYTE_SIZE);\
	add r0_index  , W, #(pr0_index*32*REG_BYTE_SIZE);\
\
	bl bs_sha256_RW_neon;


#define CALL_BS_SHA256_STEP_AVX2(H_index, E_index, G_index, F_index, u32_const, W_index, D_index, A_index, B_index, C_index) \
\
	add step_A, bs_A, #(A_index*32*REG_BYTE_SIZE);\
	add step_B, bs_A, #(B_index*32*REG_BYTE_SIZE);\
	add step_C, bs_A, #(C_index*32*REG_BYTE_SIZE);\
	add step_D, bs_A, #(D_index*32*REG_BYTE_SIZE);\
	add step_E, bs_A, #(E_index*32*REG_BYTE_SIZE);\
	add step_F, bs_A, #(F_index*32*REG_BYTE_SIZE);\
	add step_G, bs_A, #(G_index*32*REG_BYTE_SIZE);\
	add step_H, bs_A, #(H_index*32*REG_BYTE_SIZE);\
\
	movw	step_const, #(u32_const & 0xffff);\
	movt	step_const, #(u32_const >> 16);\
\
	bl bs_sha256_step_neon;


#define result  q3
#define carries q4

#define BS_ADD32_AVX2_UNROLL_0 \
		vld1.u32 {sum0}, [r8:128];\
		vld1.u32 {sum1}, [r9:128]!;\
		veor result, sum0, sum1;\
		vst1.u32 {result}, [r8:128]!;\
		vand carries, sum1, sum0;
		
#define BS_ADD32_AVX2_UNROLL(i) \
		vld1.u32 {sum0}, [r8:128];\
		vld1.u32 {sum1}, [r9:128]!;\
		veor carries, sum0, carries;\
		veor result , sum1, carries;\
		vst1.u32 {result}, [r8:128]!;\
		vbsl carries, sum1, sum0;
		
#define BS_ADD32_AVX2_UNROLL_LAST \
		vld1.u32 {sum0}, [r8:128];\
		vld1.u32 {sum1}, [r9:128]!;\
		veor carries, sum0, carries;\
		veor result , sum1, carries;\
		vst1.u32 {result}, [r8:128]!;

	.text
	.align	2
	.global	bs_add32_neon
	.type	bs_add32_neon, %function
bs_add32_neon:

	BS_ADD32_AVX2_UNROLL_0
	BS_ADD32_AVX2_UNROLL( 1)
	BS_ADD32_AVX2_UNROLL( 2)
	BS_ADD32_AVX2_UNROLL( 3)
	BS_ADD32_AVX2_UNROLL( 4)
	BS_ADD32_AVX2_UNROLL( 5)
	BS_ADD32_AVX2_UNROLL( 6)
	BS_ADD32_AVX2_UNROLL( 7)
	BS_ADD32_AVX2_UNROLL( 8)
	BS_ADD32_AVX2_UNROLL( 9)
	BS_ADD32_AVX2_UNROLL(10)
	BS_ADD32_AVX2_UNROLL(11)
	BS_ADD32_AVX2_UNROLL(12)
	BS_ADD32_AVX2_UNROLL(13)
	BS_ADD32_AVX2_UNROLL(14)
	BS_ADD32_AVX2_UNROLL(15)
	BS_ADD32_AVX2_UNROLL(16)
	BS_ADD32_AVX2_UNROLL(17)
	BS_ADD32_AVX2_UNROLL(18)
	BS_ADD32_AVX2_UNROLL(19)
	BS_ADD32_AVX2_UNROLL(20)
	BS_ADD32_AVX2_UNROLL(21)
	BS_ADD32_AVX2_UNROLL(22)
	BS_ADD32_AVX2_UNROLL(23)
	BS_ADD32_AVX2_UNROLL(24)
	BS_ADD32_AVX2_UNROLL(25)
	BS_ADD32_AVX2_UNROLL(26)
	BS_ADD32_AVX2_UNROLL(27)
	BS_ADD32_AVX2_UNROLL(28)
	BS_ADD32_AVX2_UNROLL(29)
	BS_ADD32_AVX2_UNROLL(30)
	BS_ADD32_AVX2_UNROLL_LAST
	
	bx lr


#define H 7
#define G 6
#define F 5
#define E 4
#define D 3
#define C 2
#define B 1
#define A 0

	.text
	.align	2
	.global	bs_sha256_process_block_asm_neon
	.type	bs_sha256_process_block_asm_neon, %function
bs_sha256_process_block_asm_neon:

	push  {r4,r5,r6,r7}
	push  {r8,r10,r11,r14}
	vpush {q4,q5,q6,q7}
	
	// Copy from state to tmps
	mov r8, bs_A
	mov r9, state
	mov r10, #0
copy_while:
	// Load
	vld1.u32 {q0 ,q1 }, [r9:128]!
	vld1.u32 {q2 ,q3 }, [r9:128]!
	vld1.u32 {q4 ,q5 }, [r9:128]!
	vld1.u32 {q6 ,q7 }, [r9:128]!
	vld1.u32 {q8 ,q9 }, [r9:128]!
	vld1.u32 {q10,q11}, [r9:128]!
	vld1.u32 {q12,q13}, [r9:128]!
	vld1.u32 {q14,q15}, [r9:128]!
	
	// Store
	vst1.u32 {q0 ,q1 }, [r8:128]!
	vst1.u32 {q2 ,q3 }, [r8:128]!
	vst1.u32 {q4 ,q5 }, [r8:128]!
	vst1.u32 {q6 ,q7 }, [r8:128]!
	vst1.u32 {q8 ,q9 }, [r8:128]!
	vst1.u32 {q10,q11}, [r8:128]!
	vst1.u32 {q12,q13}, [r8:128]!
	vst1.u32 {q14,q15}, [r8:128]!
	
	add r10, r10, #1
	cmp r10, #(8*32/16)
	blo copy_while
	
	vmov.u32 ones, #0xffffffff
	
    // Round 1
	mov W_ptr, W
	CALL_BS_SHA256_STEP_AVX2(H, E, G, F, 0x428A2F98,  0, D, A, B, C)
	CALL_BS_SHA256_STEP_AVX2(G, D, F, E, 0x71374491,  1, C, H, A, B)
	CALL_BS_SHA256_STEP_AVX2(F, C, E, D, 0xB5C0FBCF,  2, B, G, H, A)
	CALL_BS_SHA256_STEP_AVX2(E, B, D, C, 0xE9B5DBA5,  3, A, F, G, H)
	CALL_BS_SHA256_STEP_AVX2(D, A, C, B, 0x3956C25B,  4, H, E, F, G)
	CALL_BS_SHA256_STEP_AVX2(C, H, B, A, 0x59F111F1,  5, G, D, E, F)
	CALL_BS_SHA256_STEP_AVX2(B, G, A, H, 0x923F82A4,  6, F, C, D, E)
	CALL_BS_SHA256_STEP_AVX2(A, F, H, G, 0xAB1C5ED5,  7, E, B, C, D)
	CALL_BS_SHA256_STEP_AVX2(H, E, G, F, 0xD807AA98,  8, D, A, B, C)
	CALL_BS_SHA256_STEP_AVX2(G, D, F, E, 0x12835B01,  9, C, H, A, B)
	CALL_BS_SHA256_STEP_AVX2(F, C, E, D, 0x243185BE, 10, B, G, H, A)
	CALL_BS_SHA256_STEP_AVX2(E, B, D, C, 0x550C7DC3, 11, A, F, G, H)
	CALL_BS_SHA256_STEP_AVX2(D, A, C, B, 0x72BE5D74, 12, H, E, F, G)
	CALL_BS_SHA256_STEP_AVX2(C, H, B, A, 0x80DEB1FE, 13, G, D, E, F)
	CALL_BS_SHA256_STEP_AVX2(B, G, A, H, 0x9BDC06A7, 14, F, C, D, E)
	CALL_BS_SHA256_STEP_AVX2(A, F, H, G, 0xC19BF174, 15, E, B, C, D)
    
	// Round 2
	mov r_index, W
	add rsum_index, W, #(9*32*REG_BYTE_SIZE)
	CALL_BS_SHA256_RW_AVX2( 0, 14, 9 , 1 )
	CALL_BS_SHA256_RW_AVX2( 1, 15, 10, 2 )
	CALL_BS_SHA256_RW_AVX2( 2, 0 , 11, 3 )
	CALL_BS_SHA256_RW_AVX2( 3, 1 , 12, 4 )
	CALL_BS_SHA256_RW_AVX2( 4, 2 , 13, 5 )
	CALL_BS_SHA256_RW_AVX2( 5, 3 , 14, 6 )
	CALL_BS_SHA256_RW_AVX2( 6, 4 , 15, 7 )
	mov rsum_index, W
	CALL_BS_SHA256_RW_AVX2( 7, 5 , 0 , 8 )
	CALL_BS_SHA256_RW_AVX2( 8, 6 , 1 , 9 )
	CALL_BS_SHA256_RW_AVX2( 9, 7 , 2 , 10)
	CALL_BS_SHA256_RW_AVX2(10, 8 , 3 , 11)
	CALL_BS_SHA256_RW_AVX2(11, 9 , 4 , 12)
	CALL_BS_SHA256_RW_AVX2(12, 10, 5 , 13)
	CALL_BS_SHA256_RW_AVX2(13, 11, 6 , 14)
	CALL_BS_SHA256_RW_AVX2(14, 12, 7 , 15)
	CALL_BS_SHA256_RW_AVX2(15, 13, 8 , 0 )
	mov W_ptr, W
	CALL_BS_SHA256_STEP_AVX2(H, E, G, F, 0xE49B69C1,  0, D, A, B, C)
	CALL_BS_SHA256_STEP_AVX2(G, D, F, E, 0xEFBE4786,  1, C, H, A, B)
	CALL_BS_SHA256_STEP_AVX2(F, C, E, D, 0x0FC19DC6,  2, B, G, H, A)
	CALL_BS_SHA256_STEP_AVX2(E, B, D, C, 0x240CA1CC,  3, A, F, G, H)
	CALL_BS_SHA256_STEP_AVX2(D, A, C, B, 0x2DE92C6F,  4, H, E, F, G)
	CALL_BS_SHA256_STEP_AVX2(C, H, B, A, 0x4A7484AA,  5, G, D, E, F)
	CALL_BS_SHA256_STEP_AVX2(B, G, A, H, 0x5CB0A9DC,  6, F, C, D, E)
	CALL_BS_SHA256_STEP_AVX2(A, F, H, G, 0x76F988DA,  7, E, B, C, D)
	CALL_BS_SHA256_STEP_AVX2(H, E, G, F, 0x983E5152,  8, D, A, B, C)
	CALL_BS_SHA256_STEP_AVX2(G, D, F, E, 0xA831C66D,  9, C, H, A, B)
	CALL_BS_SHA256_STEP_AVX2(F, C, E, D, 0xB00327C8, 10, B, G, H, A)
	CALL_BS_SHA256_STEP_AVX2(E, B, D, C, 0xBF597FC7, 11, A, F, G, H)
	CALL_BS_SHA256_STEP_AVX2(D, A, C, B, 0xC6E00BF3, 12, H, E, F, G)
	CALL_BS_SHA256_STEP_AVX2(C, H, B, A, 0xD5A79147, 13, G, D, E, F)
	CALL_BS_SHA256_STEP_AVX2(B, G, A, H, 0x06CA6351, 14, F, C, D, E)
	CALL_BS_SHA256_STEP_AVX2(A, F, H, G, 0x14292967, 15, E, B, C, D)
    
	// Round 3
	mov r_index, W
	add rsum_index, W, #(9*32*REG_BYTE_SIZE)
	CALL_BS_SHA256_RW_AVX2( 0, 14, 9 , 1 )
	CALL_BS_SHA256_RW_AVX2( 1, 15, 10, 2 )
	CALL_BS_SHA256_RW_AVX2( 2, 0 , 11, 3 )
	CALL_BS_SHA256_RW_AVX2( 3, 1 , 12, 4 )
	CALL_BS_SHA256_RW_AVX2( 4, 2 , 13, 5 )
	CALL_BS_SHA256_RW_AVX2( 5, 3 , 14, 6 )
	CALL_BS_SHA256_RW_AVX2( 6, 4 , 15, 7 )
	mov rsum_index, W
	CALL_BS_SHA256_RW_AVX2( 7, 5 , 0 , 8 )
	CALL_BS_SHA256_RW_AVX2( 8, 6 , 1 , 9 )
	CALL_BS_SHA256_RW_AVX2( 9, 7 , 2 , 10)
	CALL_BS_SHA256_RW_AVX2(10, 8 , 3 , 11)
	CALL_BS_SHA256_RW_AVX2(11, 9 , 4 , 12)
	CALL_BS_SHA256_RW_AVX2(12, 10, 5 , 13)
	CALL_BS_SHA256_RW_AVX2(13, 11, 6 , 14)
	CALL_BS_SHA256_RW_AVX2(14, 12, 7 , 15)
	CALL_BS_SHA256_RW_AVX2(15, 13, 8 , 0 )
	mov W_ptr, W
	CALL_BS_SHA256_STEP_AVX2(H, E, G, F, 0x27B70A85,  0, D, A, B, C)
	CALL_BS_SHA256_STEP_AVX2(G, D, F, E, 0x2E1B2138,  1, C, H, A, B)
	CALL_BS_SHA256_STEP_AVX2(F, C, E, D, 0x4D2C6DFC,  2, B, G, H, A)
	CALL_BS_SHA256_STEP_AVX2(E, B, D, C, 0x53380D13,  3, A, F, G, H)
	CALL_BS_SHA256_STEP_AVX2(D, A, C, B, 0x650A7354,  4, H, E, F, G)
	CALL_BS_SHA256_STEP_AVX2(C, H, B, A, 0x766A0ABB,  5, G, D, E, F)
	CALL_BS_SHA256_STEP_AVX2(B, G, A, H, 0x81C2C92E,  6, F, C, D, E)
	CALL_BS_SHA256_STEP_AVX2(A, F, H, G, 0x92722C85,  7, E, B, C, D)
	CALL_BS_SHA256_STEP_AVX2(H, E, G, F, 0xA2BFE8A1,  8, D, A, B, C)
	CALL_BS_SHA256_STEP_AVX2(G, D, F, E, 0xA81A664B,  9, C, H, A, B)
	CALL_BS_SHA256_STEP_AVX2(F, C, E, D, 0xC24B8B70, 10, B, G, H, A)
	CALL_BS_SHA256_STEP_AVX2(E, B, D, C, 0xC76C51A3, 11, A, F, G, H)
	CALL_BS_SHA256_STEP_AVX2(D, A, C, B, 0xD192E819, 12, H, E, F, G)
	CALL_BS_SHA256_STEP_AVX2(C, H, B, A, 0xD6990624, 13, G, D, E, F)
	CALL_BS_SHA256_STEP_AVX2(B, G, A, H, 0xF40E3585, 14, F, C, D, E)
	CALL_BS_SHA256_STEP_AVX2(A, F, H, G, 0x106AA070, 15, E, B, C, D)
    
	// Round 4
	mov r_index, W
	add rsum_index, W, #(9*32*REG_BYTE_SIZE)
	CALL_BS_SHA256_RW_AVX2( 0, 14, 9 , 1 )
	CALL_BS_SHA256_RW_AVX2( 1, 15, 10, 2 )
	CALL_BS_SHA256_RW_AVX2( 2, 0 , 11, 3 )
	CALL_BS_SHA256_RW_AVX2( 3, 1 , 12, 4 )
	CALL_BS_SHA256_RW_AVX2( 4, 2 , 13, 5 )
	CALL_BS_SHA256_RW_AVX2( 5, 3 , 14, 6 )
	CALL_BS_SHA256_RW_AVX2( 6, 4 , 15, 7 )
	mov rsum_index, W
	CALL_BS_SHA256_RW_AVX2( 7, 5 , 0 , 8 )
	CALL_BS_SHA256_RW_AVX2( 8, 6 , 1 , 9 )
	CALL_BS_SHA256_RW_AVX2( 9, 7 , 2 , 10)
	CALL_BS_SHA256_RW_AVX2(10, 8 , 3 , 11)
	CALL_BS_SHA256_RW_AVX2(11, 9 , 4 , 12)
	CALL_BS_SHA256_RW_AVX2(12, 10, 5 , 13)
	CALL_BS_SHA256_RW_AVX2(13, 11, 6 , 14)
	CALL_BS_SHA256_RW_AVX2(14, 12, 7 , 15)
	CALL_BS_SHA256_RW_AVX2(15, 13, 8 , 0 )
	mov W_ptr, W
	CALL_BS_SHA256_STEP_AVX2(H, E, G, F, 0x19A4C116,  0, D, A, B, C)
	CALL_BS_SHA256_STEP_AVX2(G, D, F, E, 0x1E376C08,  1, C, H, A, B)
	CALL_BS_SHA256_STEP_AVX2(F, C, E, D, 0x2748774C,  2, B, G, H, A)
	CALL_BS_SHA256_STEP_AVX2(E, B, D, C, 0x34B0BCB5,  3, A, F, G, H)
	CALL_BS_SHA256_STEP_AVX2(D, A, C, B, 0x391C0CB3,  4, H, E, F, G)
	CALL_BS_SHA256_STEP_AVX2(C, H, B, A, 0x4ED8AA4A,  5, G, D, E, F)
	CALL_BS_SHA256_STEP_AVX2(B, G, A, H, 0x5B9CCA4F,  6, F, C, D, E)
	CALL_BS_SHA256_STEP_AVX2(A, F, H, G, 0x682E6FF3,  7, E, B, C, D)
	CALL_BS_SHA256_STEP_AVX2(H, E, G, F, 0x748F82EE,  8, D, A, B, C)
	CALL_BS_SHA256_STEP_AVX2(G, D, F, E, 0x78A5636F,  9, C, H, A, B)
	CALL_BS_SHA256_STEP_AVX2(F, C, E, D, 0x84C87814, 10, B, G, H, A)
	CALL_BS_SHA256_STEP_AVX2(E, B, D, C, 0x8CC70208, 11, A, F, G, H)
	CALL_BS_SHA256_STEP_AVX2(D, A, C, B, 0x90BEFFFA, 12, H, E, F, G)
	CALL_BS_SHA256_STEP_AVX2(C, H, B, A, 0xA4506CEB, 13, G, D, E, F)
	CALL_BS_SHA256_STEP_AVX2(B, G, A, H, 0xBEF9A3F7, 14, F, C, D, E)
	CALL_BS_SHA256_STEP_AVX2(A, F, H, G, 0xC67178F2, 15, E, B, C, D)
	
	mov r8, state
	mov r9, bs_A
	bl bs_add32_neon
	bl bs_add32_neon
	bl bs_add32_neon
	bl bs_add32_neon
	bl bs_add32_neon
	bl bs_add32_neon
	bl bs_add32_neon
	bl bs_add32_neon
  
	vpop {q4,q5,q6,q7}
	pop	 {r8,r10,r11,r14}	
	pop	 {r4,r5,r6,r7}
	bx lr
	