; Proof-of-concept bitslice implementation of the SHA256 compression function.
;
; Written by Alain Espinosa <alainesp at gmail.com> in 2015, and placed in the
; public domain. There's absolutely no warranty. Although this is not required
; formally, due credit will be appreciated if you re-use this code or concepts.
            
.code
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
; Bitslice SHA256 format
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
bs_A  EQU rax
state EQU rcx
W     EQU rdx
REG_BYTE_SIZE = 32


step_H     EQU r8
step_E     EQU r9
step_G     EQU r10
step_F     EQU r11
step_const EQU r12d
W_ptr      EQU r13
step_D     EQU r14
step_A     EQU r15
step_B     EQU rbx
step_C     EQU rdi

; Sums carries
carries_bs    EQU ymm4
carries_RE    EQU ymm5
carries_const EQU ymm6
carries_W     EQU ymm7
		          
carries_D     EQU ymm8
			      
carries_maj   EQU ymm9
carries_RA    EQU ymm10
ones          EQU ymm11
t0			  EQU ymm12
t1			  EQU ymm13
t2			  EQU ymm14
t3			  EQU ymm15

sum0   EQU ymm0
sum1   EQU ymm1
p      EQU ymm2
reg_H  EQU ymm3

BS_SHA256_STEP_AVX2_UNROLL MACRO i
	LOCAL zero_const, const_end
	IF (i EQ 0)
		; Calculate BITSELECT
		vmovaps     t0  , [step_G][i*REG_BYTE_SIZE]
		vpxor sum1, t0  , [step_F][i*REG_BYTE_SIZE]
		vpand sum1, sum1, [step_E][i*REG_BYTE_SIZE]
		vpxor sum1, sum1, t0
		; Sum BITSELECT
		vpxor reg_H     , sum1, [step_H][i*REG_BYTE_SIZE]
		vpand carries_bs, sum1, [step_H][i*REG_BYTE_SIZE]
		; Sum RE
		vmovaps sum0    , [step_E][((32 - 26 + i) AND 31)*REG_BYTE_SIZE]
		vpxor sum0, sum0, [step_E][((32 - 21 + i) AND 31)*REG_BYTE_SIZE]
		vpxor sum0, sum0, [step_E][((32 - 7  + i) AND 31)*REG_BYTE_SIZE]
		vpand carries_RE, sum0, reg_H
		vpxor reg_H     , sum0, reg_H
		; Sum CONST
		vpxor t1, t1, t1
		bt step_const, i
		jae zero_const
		vmovaps t1, ones
zero_const:
		vpand carries_const, t1, reg_H
		vpxor reg_H        , t1, reg_H
		; Sum W
		vpand carries_W, reg_H, [W_ptr][i*REG_BYTE_SIZE]
		vpxor reg_H    , reg_H, [W_ptr][i*REG_BYTE_SIZE]

		; Sum D
		vpxor t0       , reg_H, [step_D][i*REG_BYTE_SIZE]
		vpand carries_D, reg_H, [step_D][i*REG_BYTE_SIZE]
		vmovaps [step_D][i*REG_BYTE_SIZE], t0

		; Calculate MAJ
		vmovaps      t0 , [step_A][i*REG_BYTE_SIZE]
		vpor  sum1,  t0 , [step_B][i*REG_BYTE_SIZE]
		vpand  p  ,  t0 , [step_B][i*REG_BYTE_SIZE]
		vpand sum1, sum1, [step_C][i*REG_BYTE_SIZE]
		vpor  sum1, sum1, p
		; Sum MAJ
		vpand carries_maj, reg_H, sum1
		vpxor reg_H      , reg_H, sum1
		; Sum RA
		vmovaps sum0    , [step_A][((32 - 30 + i) AND 31)*REG_BYTE_SIZE]
		vpxor sum0, sum0, [step_A][((32 - 19 + i) AND 31)*REG_BYTE_SIZE]
		vpxor sum0, sum0, [step_A][((32 - 10 + i) AND 31)*REG_BYTE_SIZE]
		vpxor t0        , sum0, reg_H
		vpand carries_RA, sum0, reg_H
		vmovaps [step_H][i*REG_BYTE_SIZE], t0
	ELSEIF (i EQ 31)
		; Calculate BITSELECT
		vmovaps     t0  , [step_G][i*REG_BYTE_SIZE]
		vpxor sum1, t0  , [step_F][i*REG_BYTE_SIZE]
		vpand sum1, sum1, [step_E][i*REG_BYTE_SIZE]
		vpxor sum1, sum1, t0
		; Sum BITSELECT
		vpxor reg_H, sum1, [step_H][i*REG_BYTE_SIZE]
		vpxor reg_H, reg_H, carries_bs
		; Sum RE
		vpxor reg_H, reg_H, [step_E][((32 - 26 + i) AND 31)*REG_BYTE_SIZE]
		vpxor reg_H, reg_H, [step_E][((32 - 21 + i) AND 31)*REG_BYTE_SIZE]
		vpxor reg_H, reg_H, [step_E][((32 - 7  + i) AND 31)*REG_BYTE_SIZE]
		vpxor reg_H, reg_H, carries_RE
		; Sum CONST
		bt step_const, i
		jae zero_const
		vpxor reg_H, reg_H, ones
zero_const:
		vpxor reg_H, reg_H, carries_const
		; Sum W
		vpxor reg_H, reg_H, [W_ptr][i*REG_BYTE_SIZE]
		vpxor reg_H, reg_H, carries_W

		; Sum D
		vpxor p   , reg_H, [step_D][i*REG_BYTE_SIZE]
		vpxor t0, p, carries_D
		vmovaps [step_D][i*REG_BYTE_SIZE], t0

		; Calculate MAJ
		vmovaps      t0 , [step_A][i*REG_BYTE_SIZE]
		vpor  sum1,  t0 , [step_B][i*REG_BYTE_SIZE]
		vpand p   ,  t0 , [step_B][i*REG_BYTE_SIZE]
		vpand sum1, sum1, [step_C][i*REG_BYTE_SIZE]
		vpor  sum1, sum1, p
		; Sum MAJ
		vpxor reg_H, reg_H, sum1
		vpxor reg_H, reg_H, carries_maj
		; Sum RA
		vpxor reg_H, reg_H, [step_A][((32 - 30 + i) AND 31)*REG_BYTE_SIZE]
		vpxor reg_H, reg_H, [step_A][((32 - 19 + i) AND 31)*REG_BYTE_SIZE]
		vpxor reg_H, reg_H, [step_A][((32 - 10 + i) AND 31)*REG_BYTE_SIZE]
		vpxor reg_H, reg_H, carries_RA
		vmovaps [step_H][i*REG_BYTE_SIZE], reg_H
	ELSE
		; Calculate BITSELECT
		;vmovaps     t0  , [step_G][i*REG_BYTE_SIZE]
		;vpxor sum1, t0  , [step_F][i*REG_BYTE_SIZE]
		;vpand sum1, sum1, [step_E][i*REG_BYTE_SIZE]
		;vpxor sum1, sum1, t0		
		; Alternative BITSELECT using vpandn with better scheduling
		vmovaps       t0, [step_E][i*REG_BYTE_SIZE]
		vpandn   t1 , t0, [step_G][i*REG_BYTE_SIZE]
		vpand   sum1, t0, [step_F][i*REG_BYTE_SIZE]
		vpor    sum1, t1, sum1
		; Sum BITSELECT
		vpxor p   , sum1, [step_H][i*REG_BYTE_SIZE]
		vpxor reg_H, p  , carries_bs
		vpand sum0, sum1, [step_H][i*REG_BYTE_SIZE]
		vpand carries_bs, carries_bs, p
		vpor  carries_bs, carries_bs, sum0
		; Sum RE
		vmovaps sum0    , [step_E][((32 - 26 + i) AND 31)*REG_BYTE_SIZE]
		vpxor sum0, sum0, [step_E][((32 - 21 + i) AND 31)*REG_BYTE_SIZE]
		vpxor sum0, sum0, [step_E][((32 - 7  + i) AND 31)*REG_BYTE_SIZE]
		vpxor p   , sum0, reg_H
		vpand sum0, sum0, reg_H
		vpxor reg_H  , p, carries_RE
		vpand carries_RE, carries_RE, p
		vpor  carries_RE, carries_RE, sum0
		; Sum CONST
		bt step_const, i
		jae zero_const; ONES
			vpxor t1, reg_H, carries_const
			vpor  carries_const, t1, reg_H
			vpxor reg_H, t1, ones
			jmp const_end
zero_const:; ZERO
			vpxor reg_H, reg_H, carries_const
			vpandn carries_const, reg_H, carries_const
const_end:
		; Sum W
		vpxor p   , reg_H, [W_ptr][i*REG_BYTE_SIZE]
		vpand sum0, reg_H, [W_ptr][i*REG_BYTE_SIZE]
		vpxor reg_H , p, carries_W
		vpand carries_W, carries_W, p
		vpor  carries_W, carries_W, sum0

		; Sum D
		vpxor p   , reg_H, [step_D][i*REG_BYTE_SIZE]
		vpand sum0, reg_H, [step_D][i*REG_BYTE_SIZE]
		vpxor t0, p, carries_D
		vmovaps [step_D][i*REG_BYTE_SIZE], t0
		vpand carries_D, carries_D, p
		vpor  carries_D, carries_D, sum0

		; Calculate MAJ
		vmovaps      t0 , [step_A][i*REG_BYTE_SIZE]
		vpor  sum1,  t0 , [step_B][i*REG_BYTE_SIZE]
		vpand p   ,  t0 , [step_B][i*REG_BYTE_SIZE]
		vpand sum1, sum1, [step_C][i*REG_BYTE_SIZE]
		vpor  sum1, sum1, p
		; Sum MAJ
		vpxor p   , reg_H, sum1
		vpand sum1, reg_H, sum1
		vpxor reg_H, p, carries_maj
		vpand carries_maj, carries_maj, p
		vpor  carries_maj, carries_maj, sum1
		; Sum RA
		vmovaps     sum0, [step_A][((32 - 30 + i) AND 31)*REG_BYTE_SIZE]
		vpxor sum0, sum0, [step_A][((32 - 19 + i) AND 31)*REG_BYTE_SIZE]
		vpxor sum0, sum0, [step_A][((32 - 10 + i) AND 31)*REG_BYTE_SIZE]
		vpxor p   , sum0, reg_H
		vpand sum0, sum0, reg_H
		vpxor t0, p, carries_RA
		vmovaps [step_H][i*REG_BYTE_SIZE], t0
		vpand carries_RA, carries_RA, p
		vpor  carries_RA, carries_RA, sum0
	ENDIF
ENDM


bs_sha256_step_avx2 PROC
	vpcmpeqd ones, ones, ones 
	
	FOR bit_index, <0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31>
		BS_SHA256_STEP_AVX2_UNROLL bit_index
	ENDM
	
	RET
bs_sha256_step_avx2 ENDP

; Sums carries
carries_R1  EQU ymm4
carries_sum EQU ymm5
carries_R0  EQU ymm6

sum0   EQU ymm0
sum1   EQU ymm1
p      EQU ymm2
reg_W  EQU ymm3

r_index    EQU r8
r1_index   EQU r9
rsum_index EQU r10
r0_index   EQU r11

BS_SHA256_RW_UNROLL MACRO i
	IF (i EQ 0)
		; Calculate R1
		vmovaps sum1    , [r1_index][((32 - 15 + i) AND 31)*REG_BYTE_SIZE]
		vpxor sum1, sum1, [r1_index][((32 - 13 + i) AND 31)*REG_BYTE_SIZE]
		vpxor sum1, sum1, [r1_index][(10 + i)*REG_BYTE_SIZE]
		; Sum R1
		vpxor reg_W     , sum1, [r_index][i*REG_BYTE_SIZE]
		vpand carries_R1, sum1, [r_index][i*REG_BYTE_SIZE]

		; Sum rsum_index
		vpand carries_sum, reg_W, [rsum_index][i*REG_BYTE_SIZE]
		vpxor reg_W      , reg_W, [rsum_index][i*REG_BYTE_SIZE]

		; Calculate R0
		vmovaps sum0    , [r0_index][((32 - 25 + i) AND 31)*REG_BYTE_SIZE]
		vpxor sum0, sum0, [r0_index][((32 - 14 + i) AND 31)*REG_BYTE_SIZE]
		vpxor sum0, sum0, [r0_index][(3 + i)*REG_BYTE_SIZE]
		; Sum R0
		vpand carries_R0, sum0, reg_W
		vpxor reg_W     , sum0, reg_W
		vmovaps [r_index][i*REG_BYTE_SIZE], reg_W
	ELSEIF (i EQ 31)
		; Calculate R1
		vmovaps reg_W     , [r1_index][((32 - 15 + i) AND 31)*REG_BYTE_SIZE]
		vpxor reg_W, reg_W, [r1_index][((32 - 13 + i) AND 31)*REG_BYTE_SIZE]
		; Sum R1
		vpxor reg_W, reg_W, [r_index][i*REG_BYTE_SIZE]
		vpxor reg_W, reg_W, carries_R1

		; Sum rsum_index
		vpxor reg_W, reg_W, [rsum_index][i*REG_BYTE_SIZE]
		vpxor reg_W, reg_W, carries_sum

		; Calculate R0
		vpxor reg_W, reg_W, [r0_index][((32 - 25 + i) AND 31)*REG_BYTE_SIZE]
		vpxor reg_W, reg_W, [r0_index][((32 - 14 + i) AND 31)*REG_BYTE_SIZE]
		; Sum R0
		vpxor reg_W, reg_W, carries_R0
		vmovaps [r_index][i*REG_BYTE_SIZE], reg_W
	ELSE
		; Calculate R1
		vmovaps sum1    , [r1_index][((32 - 15 + i) AND 31)*REG_BYTE_SIZE]
		vpxor sum1, sum1, [r1_index][((32 - 13 + i) AND 31)*REG_BYTE_SIZE]
		IF (i LT (32-10))
			vpxor sum1, sum1, [r1_index][(10 + i)*REG_BYTE_SIZE]
		ENDIF
		; Sum R1
		vpxor p   , sum1, [r_index][i*REG_BYTE_SIZE]
		vpxor reg_W, p, carries_R1
		vpand sum0, sum1, [r_index][i*REG_BYTE_SIZE]
		vpand carries_R1, carries_R1, p
		vpor carries_R1, carries_R1, sum0

		; Sum rsum_index
		vpxor p   , reg_W, [rsum_index][i*REG_BYTE_SIZE]
		vpand sum0, reg_W, [rsum_index][i*REG_BYTE_SIZE]
		vpxor reg_W, p, carries_sum
		vpand carries_sum, carries_sum, p
		vpor carries_sum, carries_sum, sum0

		; Calculate R0
		vmovaps sum0    , [r0_index][((32 - 25 + i) AND 31)*REG_BYTE_SIZE]
		vpxor sum0, sum0, [r0_index][((32 - 14 + i) AND 31)*REG_BYTE_SIZE]
		IF (i LT (32-3))
			vpxor sum0, sum0, [r0_index][(3 + i)*REG_BYTE_SIZE]
		ENDIF
		; Sum R0
		vpxor p, sum0, reg_W
		vpand sum0, sum0, reg_W
		vpxor reg_W, p, carries_R0
		vmovaps [r_index][i*REG_BYTE_SIZE], reg_W
		vpand carries_R0, carries_R0, p
		vpor carries_R0, carries_R0, sum0
	ENDIF
ENDM

bs_sha256_RW_avx2 PROC
	FOR bit_index, <0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31>
		BS_SHA256_RW_UNROLL bit_index
	ENDM
	RET
bs_sha256_RW_avx2 ENDP

RECALCULATE_RW_POINTERS MACRO rw_ptr, index_ptr
	IF (index_ptr EQ 0)
		sub rw_ptr, 15*32*REG_BYTE_SIZE
	ELSE
		add rw_ptr, 32*REG_BYTE_SIZE
	ENDIF
ENDM

CALL_BS_SHA256_RW_AVX2 MACRO pr_index, pr1_index, prsum_index, pr0_index

	IF (pr_index EQ 0)
		mov r_index, W
	ELSE
		add r_index, 32*REG_BYTE_SIZE
	ENDIF

	RECALCULATE_RW_POINTERS r1_index  , pr1_index  
	RECALCULATE_RW_POINTERS rsum_index, prsum_index
	RECALCULATE_RW_POINTERS r0_index  , pr0_index  
	
	call bs_sha256_RW_avx2
ENDM

RECALCULATE_STEP_POINTERS MACRO step_ptr, index_ptr
	IF (index_ptr EQ 7)
		add step_ptr, 7*32*REG_BYTE_SIZE
	ELSE
		sub step_ptr, 32*REG_BYTE_SIZE
	ENDIF
ENDM

CALL_BS_SHA256_STEP_AVX2 MACRO H_index, E_index, G_index, F_index, u32_const, W_index, D_index, A_index, B_index, C_index

	IF (W_index EQ 0)
		mov W_ptr, W
	ELSE
		add W_ptr, 32*REG_BYTE_SIZE
	ENDIF

	RECALCULATE_STEP_POINTERS step_A, A_index
	RECALCULATE_STEP_POINTERS step_B, B_index
	RECALCULATE_STEP_POINTERS step_C, C_index
	RECALCULATE_STEP_POINTERS step_D, D_index
	RECALCULATE_STEP_POINTERS step_E, E_index
	RECALCULATE_STEP_POINTERS step_F, F_index
	RECALCULATE_STEP_POINTERS step_G, G_index
	RECALCULATE_STEP_POINTERS step_H, H_index

	mov step_const, u32_const
	
	call bs_sha256_step_avx2
ENDM


result  EQU ymm3
carries EQU ymm4

BS_ADD32_AVX2_UNROLL MACRO i
	IF (i EQ 0)
		vmovaps carries, [r8][i*REG_BYTE_SIZE]
		vmovaps sum1   , [r9][i*REG_BYTE_SIZE]
		vpxor result , carries, sum1
		vpand carries, carries, sum1
		vmovaps [r8][i*REG_BYTE_SIZE], result
	ELSEIF (i EQ 31)
		vmovaps sum0, [r8][i*REG_BYTE_SIZE]
		vmovaps sum1, [r9][i*REG_BYTE_SIZE]
		vpxor p, sum0, sum1
		vpxor result, p, carries
		vmovaps [r8][i*REG_BYTE_SIZE], result
	ElSE
		vmovaps sum0, [r8][i*REG_BYTE_SIZE]
		vmovaps sum1, [r9][i*REG_BYTE_SIZE]
		vpxor p, sum0, sum1
		vpxor result, p, carries
		vpand sum0, sum0, sum1
		vpand carries, carries, p
		vmovaps [r8][i*REG_BYTE_SIZE], result
		vpor carries, carries, sum0
	ENDIF
ENDM

bs_add32_avx2 PROC
	FOR bit_index, <0,1,2,3,4,5,6,7,8,9,10,11,12,13,14,15,16,17,18,19,20,21,22,23,24,25,26,27,28,29,30,31>
		BS_ADD32_AVX2_UNROLL bit_index
	ENDM
	RET
bs_add32_avx2 ENDP

H EQU 7
G EQU 6
F EQU 5
E EQU 4
D EQU 3
C EQU 2
B EQU 1
A EQU 0

bs_sha256_process_block_asm_avx2 PROC
	push r12
	push r13
	push r14
	push r15
	push rbx
	push rdi
	
	mov bs_A, r8
	
	; Copy from state to tmps
	; r8 point to tmps
	mov r9, state
	mov r10, 8*32/16
copy_while:
	; Load
	vmovaps ymm0 , [r9+0 *REG_BYTE_SIZE]
	vmovaps ymm1 , [r9+1 *REG_BYTE_SIZE]
	vmovaps ymm2 , [r9+2 *REG_BYTE_SIZE]
	vmovaps ymm3 , [r9+3 *REG_BYTE_SIZE]
	vmovaps ymm4 , [r9+4 *REG_BYTE_SIZE]
	vmovaps ymm5 , [r9+5 *REG_BYTE_SIZE]
	vmovaps ymm6 , [r9+6 *REG_BYTE_SIZE]
	vmovaps ymm7 , [r9+7 *REG_BYTE_SIZE]
	vmovaps ymm8 , [r9+8 *REG_BYTE_SIZE]
	vmovaps ymm9 , [r9+9 *REG_BYTE_SIZE]
	vmovaps ymm10, [r9+10*REG_BYTE_SIZE]
	vmovaps ymm11, [r9+11*REG_BYTE_SIZE]
	vmovaps ymm12, [r9+12*REG_BYTE_SIZE]
	vmovaps ymm13, [r9+13*REG_BYTE_SIZE]
	vmovaps ymm14, [r9+14*REG_BYTE_SIZE]
	vmovaps ymm15, [r9+15*REG_BYTE_SIZE]
	
	; Store
	vmovaps [r8+0 *REG_BYTE_SIZE], ymm0 
	vmovaps [r8+1 *REG_BYTE_SIZE], ymm1 
	vmovaps [r8+2 *REG_BYTE_SIZE], ymm2 
	vmovaps [r8+3 *REG_BYTE_SIZE], ymm3 
	vmovaps [r8+4 *REG_BYTE_SIZE], ymm4 
	vmovaps [r8+5 *REG_BYTE_SIZE], ymm5 
	vmovaps [r8+6 *REG_BYTE_SIZE], ymm6 
	vmovaps [r8+7 *REG_BYTE_SIZE], ymm7 
	vmovaps [r8+8 *REG_BYTE_SIZE], ymm8 
	vmovaps [r8+9 *REG_BYTE_SIZE], ymm9 
	vmovaps [r8+10*REG_BYTE_SIZE], ymm10
	vmovaps [r8+11*REG_BYTE_SIZE], ymm11
	vmovaps [r8+12*REG_BYTE_SIZE], ymm12
	vmovaps [r8+13*REG_BYTE_SIZE], ymm13
	vmovaps [r8+14*REG_BYTE_SIZE], ymm14
	vmovaps [r8+15*REG_BYTE_SIZE], ymm15
	
	add r8, 16*REG_BYTE_SIZE
	add r9, 16*REG_BYTE_SIZE
	dec r10
	jnz copy_while
	
	; Initialize pointers
	mov step_A, bs_A
	mov step_B, bs_A
	mov step_C, bs_A
	mov step_D, bs_A
	add step_A, 1*32*REG_BYTE_SIZE
	add step_B, 2*32*REG_BYTE_SIZE
	add step_C, 3*32*REG_BYTE_SIZE
	add step_D, 4*32*REG_BYTE_SIZE
	
	mov step_E, bs_A
	mov step_F, bs_A
	mov step_G, bs_A
	mov step_H, bs_A
	add step_E, 5*32*REG_BYTE_SIZE
	add step_F, 6*32*REG_BYTE_SIZE
	add step_G, 7*32*REG_BYTE_SIZE
  
    ; Round 1
	CALL_BS_SHA256_STEP_AVX2 H, E, G, F, 0428A2F98H,  0, D, A, B, C
	CALL_BS_SHA256_STEP_AVX2 G, D, F, E, 071374491H,  1, C, H, A, B
	CALL_BS_SHA256_STEP_AVX2 F, C, E, D, 0B5C0FBCFH,  2, B, G, H, A
	CALL_BS_SHA256_STEP_AVX2 E, B, D, C, 0E9B5DBA5H,  3, A, F, G, H
	CALL_BS_SHA256_STEP_AVX2 D, A, C, B, 03956C25BH,  4, H, E, F, G
	CALL_BS_SHA256_STEP_AVX2 C, H, B, A, 059F111F1H,  5, G, D, E, F
	CALL_BS_SHA256_STEP_AVX2 B, G, A, H, 0923F82A4H,  6, F, C, D, E
	CALL_BS_SHA256_STEP_AVX2 A, F, H, G, 0AB1C5ED5H,  7, E, B, C, D
	CALL_BS_SHA256_STEP_AVX2 H, E, G, F, 0D807AA98H,  8, D, A, B, C
	CALL_BS_SHA256_STEP_AVX2 G, D, F, E, 012835B01H,  9, C, H, A, B
	CALL_BS_SHA256_STEP_AVX2 F, C, E, D, 0243185BEH, 10, B, G, H, A
	CALL_BS_SHA256_STEP_AVX2 E, B, D, C, 0550C7DC3H, 11, A, F, G, H
	CALL_BS_SHA256_STEP_AVX2 D, A, C, B, 072BE5D74H, 12, H, E, F, G
	CALL_BS_SHA256_STEP_AVX2 C, H, B, A, 080DEB1FEH, 13, G, D, E, F
	CALL_BS_SHA256_STEP_AVX2 B, G, A, H, 09BDC06A7H, 14, F, C, D, E
	CALL_BS_SHA256_STEP_AVX2 A, F, H, G, 0C19BF174H, 15, E, B, C, D
    
	; Round 2
	mov r1_index  , W
	mov rsum_index, W
	mov r0_index  , W
	add r1_index  , 13*32*REG_BYTE_SIZE
	add rsum_index,  8*32*REG_BYTE_SIZE
	CALL_BS_SHA256_RW_AVX2  0, 14, 9 , 1 
	CALL_BS_SHA256_RW_AVX2  1, 15, 10, 2 
	CALL_BS_SHA256_RW_AVX2  2, 0 , 11, 3 
	CALL_BS_SHA256_RW_AVX2  3, 1 , 12, 4 
	CALL_BS_SHA256_RW_AVX2  4, 2 , 13, 5 
	CALL_BS_SHA256_RW_AVX2  5, 3 , 14, 6 
	CALL_BS_SHA256_RW_AVX2  6, 4 , 15, 7 
	CALL_BS_SHA256_RW_AVX2  7, 5 , 0 , 8 
	CALL_BS_SHA256_RW_AVX2  8, 6 , 1 , 9 
	CALL_BS_SHA256_RW_AVX2  9, 7 , 2 , 10
	CALL_BS_SHA256_RW_AVX2 10, 8 , 3 , 11
	CALL_BS_SHA256_RW_AVX2 11, 9 , 4 , 12
	CALL_BS_SHA256_RW_AVX2 12, 10, 5 , 13
	CALL_BS_SHA256_RW_AVX2 13, 11, 6 , 14
	CALL_BS_SHA256_RW_AVX2 14, 12, 7 , 15
	CALL_BS_SHA256_RW_AVX2 15, 13, 8 , 0 
	mov step_E, bs_A
	mov step_F, bs_A
	mov step_G, bs_A
	mov step_H, bs_A
	add step_E, 5*32*REG_BYTE_SIZE
	add step_F, 6*32*REG_BYTE_SIZE
	add step_G, 7*32*REG_BYTE_SIZE
	CALL_BS_SHA256_STEP_AVX2 H, E, G, F, 0E49B69C1H,  0, D, A, B, C
	CALL_BS_SHA256_STEP_AVX2 G, D, F, E, 0EFBE4786H,  1, C, H, A, B
	CALL_BS_SHA256_STEP_AVX2 F, C, E, D, 00FC19DC6H,  2, B, G, H, A
	CALL_BS_SHA256_STEP_AVX2 E, B, D, C, 0240CA1CCH,  3, A, F, G, H
	CALL_BS_SHA256_STEP_AVX2 D, A, C, B, 02DE92C6FH,  4, H, E, F, G
	CALL_BS_SHA256_STEP_AVX2 C, H, B, A, 04A7484AAH,  5, G, D, E, F
	CALL_BS_SHA256_STEP_AVX2 B, G, A, H, 05CB0A9DCH,  6, F, C, D, E
	CALL_BS_SHA256_STEP_AVX2 A, F, H, G, 076F988DAH,  7, E, B, C, D
	CALL_BS_SHA256_STEP_AVX2 H, E, G, F, 0983E5152H,  8, D, A, B, C
	CALL_BS_SHA256_STEP_AVX2 G, D, F, E, 0A831C66DH,  9, C, H, A, B
	CALL_BS_SHA256_STEP_AVX2 F, C, E, D, 0B00327C8H, 10, B, G, H, A
	CALL_BS_SHA256_STEP_AVX2 E, B, D, C, 0BF597FC7H, 11, A, F, G, H
	CALL_BS_SHA256_STEP_AVX2 D, A, C, B, 0C6E00BF3H, 12, H, E, F, G
	CALL_BS_SHA256_STEP_AVX2 C, H, B, A, 0D5A79147H, 13, G, D, E, F
	CALL_BS_SHA256_STEP_AVX2 B, G, A, H, 006CA6351H, 14, F, C, D, E
	CALL_BS_SHA256_STEP_AVX2 A, F, H, G, 014292967H, 15, E, B, C, D

	; Round 3
	mov r1_index  , W
	mov rsum_index, W
	mov r0_index  , W
	add r1_index  , 13*32*REG_BYTE_SIZE
	add rsum_index,  8*32*REG_BYTE_SIZE
	CALL_BS_SHA256_RW_AVX2  0, 14, 9 , 1 
	CALL_BS_SHA256_RW_AVX2  1, 15, 10, 2 
	CALL_BS_SHA256_RW_AVX2  2, 0 , 11, 3 
	CALL_BS_SHA256_RW_AVX2  3, 1 , 12, 4 
	CALL_BS_SHA256_RW_AVX2  4, 2 , 13, 5 
	CALL_BS_SHA256_RW_AVX2  5, 3 , 14, 6 
	CALL_BS_SHA256_RW_AVX2  6, 4 , 15, 7 
	CALL_BS_SHA256_RW_AVX2  7, 5 , 0 , 8 
	CALL_BS_SHA256_RW_AVX2  8, 6 , 1 , 9 
	CALL_BS_SHA256_RW_AVX2  9, 7 , 2 , 10
	CALL_BS_SHA256_RW_AVX2 10, 8 , 3 , 11
	CALL_BS_SHA256_RW_AVX2 11, 9 , 4 , 12
	CALL_BS_SHA256_RW_AVX2 12, 10, 5 , 13
	CALL_BS_SHA256_RW_AVX2 13, 11, 6 , 14
	CALL_BS_SHA256_RW_AVX2 14, 12, 7 , 15
	CALL_BS_SHA256_RW_AVX2 15, 13, 8 , 0 
	mov step_E, bs_A
	mov step_F, bs_A
	mov step_G, bs_A
	mov step_H, bs_A
	add step_E, 5*32*REG_BYTE_SIZE
	add step_F, 6*32*REG_BYTE_SIZE
	add step_G, 7*32*REG_BYTE_SIZE
	CALL_BS_SHA256_STEP_AVX2 H, E, G, F, 027B70A85H,  0, D, A, B, C
	CALL_BS_SHA256_STEP_AVX2 G, D, F, E, 02E1B2138H,  1, C, H, A, B
	CALL_BS_SHA256_STEP_AVX2 F, C, E, D, 04D2C6DFCH,  2, B, G, H, A
	CALL_BS_SHA256_STEP_AVX2 E, B, D, C, 053380D13H,  3, A, F, G, H
	CALL_BS_SHA256_STEP_AVX2 D, A, C, B, 0650A7354H,  4, H, E, F, G
	CALL_BS_SHA256_STEP_AVX2 C, H, B, A, 0766A0ABBH,  5, G, D, E, F
	CALL_BS_SHA256_STEP_AVX2 B, G, A, H, 081C2C92EH,  6, F, C, D, E
	CALL_BS_SHA256_STEP_AVX2 A, F, H, G, 092722C85H,  7, E, B, C, D
	CALL_BS_SHA256_STEP_AVX2 H, E, G, F, 0A2BFE8A1H,  8, D, A, B, C
	CALL_BS_SHA256_STEP_AVX2 G, D, F, E, 0A81A664BH,  9, C, H, A, B
	CALL_BS_SHA256_STEP_AVX2 F, C, E, D, 0C24B8B70H, 10, B, G, H, A
	CALL_BS_SHA256_STEP_AVX2 E, B, D, C, 0C76C51A3H, 11, A, F, G, H
	CALL_BS_SHA256_STEP_AVX2 D, A, C, B, 0D192E819H, 12, H, E, F, G
	CALL_BS_SHA256_STEP_AVX2 C, H, B, A, 0D6990624H, 13, G, D, E, F
	CALL_BS_SHA256_STEP_AVX2 B, G, A, H, 0F40E3585H, 14, F, C, D, E
	CALL_BS_SHA256_STEP_AVX2 A, F, H, G, 0106AA070H, 15, E, B, C, D

	; Round 4
	mov r1_index  , W
	mov rsum_index, W
	mov r0_index  , W
	add r1_index  , 13*32*REG_BYTE_SIZE
	add rsum_index,  8*32*REG_BYTE_SIZE
	CALL_BS_SHA256_RW_AVX2  0, 14, 9 , 1 
	CALL_BS_SHA256_RW_AVX2  1, 15, 10, 2 
	CALL_BS_SHA256_RW_AVX2  2, 0 , 11, 3 
	CALL_BS_SHA256_RW_AVX2  3, 1 , 12, 4 
	CALL_BS_SHA256_RW_AVX2  4, 2 , 13, 5 
	CALL_BS_SHA256_RW_AVX2  5, 3 , 14, 6 
	CALL_BS_SHA256_RW_AVX2  6, 4 , 15, 7 
	CALL_BS_SHA256_RW_AVX2  7, 5 , 0 , 8 
	CALL_BS_SHA256_RW_AVX2  8, 6 , 1 , 9 
	CALL_BS_SHA256_RW_AVX2  9, 7 , 2 , 10
	CALL_BS_SHA256_RW_AVX2 10, 8 , 3 , 11
	CALL_BS_SHA256_RW_AVX2 11, 9 , 4 , 12
	CALL_BS_SHA256_RW_AVX2 12, 10, 5 , 13
	CALL_BS_SHA256_RW_AVX2 13, 11, 6 , 14
	CALL_BS_SHA256_RW_AVX2 14, 12, 7 , 15
	CALL_BS_SHA256_RW_AVX2 15, 13, 8 , 0 
	mov step_E, bs_A
	mov step_F, bs_A
	mov step_G, bs_A
	mov step_H, bs_A
	add step_E, 5*32*REG_BYTE_SIZE
	add step_F, 6*32*REG_BYTE_SIZE
	add step_G, 7*32*REG_BYTE_SIZE
	CALL_BS_SHA256_STEP_AVX2 H, E, G, F, 019A4C116H,  0, D, A, B, C
	CALL_BS_SHA256_STEP_AVX2 G, D, F, E, 01E376C08H,  1, C, H, A, B
	CALL_BS_SHA256_STEP_AVX2 F, C, E, D, 02748774CH,  2, B, G, H, A
	CALL_BS_SHA256_STEP_AVX2 E, B, D, C, 034B0BCB5H,  3, A, F, G, H
	CALL_BS_SHA256_STEP_AVX2 D, A, C, B, 0391C0CB3H,  4, H, E, F, G
	CALL_BS_SHA256_STEP_AVX2 C, H, B, A, 04ED8AA4AH,  5, G, D, E, F
	CALL_BS_SHA256_STEP_AVX2 B, G, A, H, 05B9CCA4FH,  6, F, C, D, E
	CALL_BS_SHA256_STEP_AVX2 A, F, H, G, 0682E6FF3H,  7, E, B, C, D
	CALL_BS_SHA256_STEP_AVX2 H, E, G, F, 0748F82EEH,  8, D, A, B, C
	CALL_BS_SHA256_STEP_AVX2 G, D, F, E, 078A5636FH,  9, C, H, A, B
	CALL_BS_SHA256_STEP_AVX2 F, C, E, D, 084C87814H, 10, B, G, H, A
	CALL_BS_SHA256_STEP_AVX2 E, B, D, C, 08CC70208H, 11, A, F, G, H
	CALL_BS_SHA256_STEP_AVX2 D, A, C, B, 090BEFFFAH, 12, H, E, F, G
	CALL_BS_SHA256_STEP_AVX2 C, H, B, A, 0A4506CEBH, 13, G, D, E, F
	CALL_BS_SHA256_STEP_AVX2 B, G, A, H, 0BEF9A3F7H, 14, F, C, D, E
	CALL_BS_SHA256_STEP_AVX2 A, F, H, G, 0C67178F2H, 15, E, B, C, D
	
	mov r8, state
	mov r9, bs_A
	call bs_add32_avx2
	
	add r8, 32*REG_BYTE_SIZE
	add r9, 32*REG_BYTE_SIZE
	call bs_add32_avx2
	
	add r8, 32*REG_BYTE_SIZE
	add r9, 32*REG_BYTE_SIZE
	call bs_add32_avx2
	
	add r8, 32*REG_BYTE_SIZE
	add r9, 32*REG_BYTE_SIZE
	call bs_add32_avx2
	
	add r8, 32*REG_BYTE_SIZE
	add r9, 32*REG_BYTE_SIZE
	call bs_add32_avx2
	
	add r8, 32*REG_BYTE_SIZE
	add r9, 32*REG_BYTE_SIZE
	call bs_add32_avx2
	
	add r8, 32*REG_BYTE_SIZE
	add r9, 32*REG_BYTE_SIZE
	call bs_add32_avx2
	
	add r8, 32*REG_BYTE_SIZE
	add r9, 32*REG_BYTE_SIZE
	call bs_add32_avx2
  
	pop rdi
	pop rbx
	pop r15
	pop r14
	pop r13
	pop r12
	
	RET
bs_sha256_process_block_asm_avx2 ENDP

end